# sigma-point bayesian processor, unscented kalman filter, SPBP, UKF
import numpy as np
import math
from innov import Innov

class Modern():
    def __init__(self):
        pass

    def spbp1(self):
        n = 150
        deltat = .01

        xhat0 = 2.2
        Ptil0 = .01

        bignsubx = 1
        kappa = 1
        bk = bignsubx + kappa
        W = np.zeros((3,))
        W[:] = [kappa / float(bk), .5 / float(bk), .5 / float(bk)]

        def fx(x, w):
            return (1 - .05 * deltat) * x + .04 * deltat * x**2 + w

        def fy(x, v):
            return x ** 2 + x ** 3 + v

        vfx = np.vectorize(fx)
        vfy = np.vectorize(fy)

        def vfa(vx):
            return vfx(vx, 0)

        def vfc(vx):
            return vfy(vx, 0)

        def vfX(xhat, Ptil):
            return [xhat, xhat + math.sqrt(bk * Ptil), xhat - math.sqrt(bk * Ptil)]

        def Xhat(X, Rww):
            return [X[0], X[1] + kappa * math.sqrt(Rww), X[2] - kappa * math.sqrt(Rww)]

        tts = np.arange(0, n * deltat, deltat)
        Rvv = .09
        vts = math.sqrt(Rvv) * np.random.randn(n)
        Rww = 0
        wts = math.sqrt(Rww) * np.random.randn(n)

        xts = np.zeros((n,))
        xts[0] = 2.
        yts = np.zeros((n,))
        yts[0] = fy(xts[0], vts[0])

        Xts = np.zeros((n, 3))
        Xts[0, :] = vfX(xhat0, Ptil0)
        xhatts = np.zeros((n,))
        xhatts[0] = xhat0
        xtilts = np.zeros((n,))
        xtilts[0] = xts[0] - xhatts[0]

        Xtilts = np.zeros((n, 3))
        Xtilts[0, :] = Xts[0, :] - xhatts[0]
        Ptilts = np.zeros((n,))
        Ptilts[0] = Ptil0

        Xhatts = np.zeros((n, 3))
        Xhatts[0, :] = Xts[0, :]

        Yts = np.zeros((n, 3))
        Yts[0, :] = vfc(Xhatts[0, :])
        yhatts = np.zeros((n,))
        yhatts[0] = W @ Yts[0, :]
        ets = np.zeros((n,))
        ets[0] = yts[0] - yhatts[0]

        ksits = np.zeros((n, 3))
        ksits[0, :] = Yts[0, :] - yhatts[0]
        Rksiksits = np.zeros((n,))
        Rksiksits[0] = W @ np.power(ksits[0, :], 2) + Rvv

        RXtilksits = np.zeros((n,))
        RXtilksits[0] = W @ np.multiply(Xtilts[0, :], ksits[0, :])
        Kts = np.zeros((n,))
        Kts[0] = RXtilksits[0] / Rksiksits[0]

        for tk in range(1, n):
            xts[tk] = fx(xts[tk - 1], wts[tk - 1])
            yts[tk] = fy(xts[tk], vts[tk])

            X = vfX(xhatts[tk-1], Ptilts[tk-1])
            Xts[tk, :] = vfa(X)
            xhatts[tk] = W @ Xts[tk, :]
            Xtilts[tk, :] = Xts[tk, :] - xhatts[tk]
            Ptilts[tk] = W @ np.power(Xtilts[tk, :], 2) + Rww

            Xhatts[tk, :] = Xhat(Xts[tk, :], Rww)

            Yts[tk, :] = vfc(Xhatts[tk, :])
            yhatts[tk] = W @ Yts[tk, :]
            ets[tk] = yts[tk] - yhatts[tk]

            ksits[tk, :] = Yts[tk, :] - yhatts[tk]
            Rksiksits[tk] = W @ np.power(ksits[tk, :], 2) + Rvv

            RXtilksits[tk] = W @ np.multiply(Xtilts[tk, :], ksits[tk, :])
            Kts[tk] = RXtilksits[tk] / Rksiksits[tk]

            xhatts[tk] = xhatts[tk] + Kts[tk] * ets[tk]
            Ptilts[tk] = Ptilts[tk] - Kts[tk] * Rksiksits[tk] * Kts[tk]
            xtilts[tk] = xts[tk] - xhatts[tk]

        innov = Innov(tts, ets)
        innov.plot_standard(tts, xhatts, xtilts, yhatts)

    def spbp2(self):
        n = 150
        deltat = .01
        tts = np.arange(0, n * deltat, deltat)

        Rww = np.diag([0, 0, 0])
        wts = np.multiply(np.random.randn(n, 3), np.sqrt(np.diag(Rww)))
        xts = np.zeros([n, 3])
        xts[0, :] = [2., .05, .04]

        def fx(x, w):
            return np.array([(1 - x[1] * deltat) * x[0] + x[2] * deltat * x[0] ** 2, x[1], x[2]]) + w

        for tk in range(1, n):
            xts[tk, :] = fx(xts[tk - 1, :], wts[tk - 1, :])

        Rvv = .09
        vts = math.sqrt(Rvv) * np.random.randn(n)
        yts = np.zeros(n)

        def fy(x, v):
            return x[0] ** 2 + x[0] ** 3 + v

        yts[0] = fy(xts[0, :], vts[0])
        for tk in range(1, n):
            yts[tk] = fy(xts[tk, :], vts[tk])

        nx, kappa = 3, 1
        a, b = kappa / float(nx + kappa), 1 / float(2 * (nx + kappa))
        W = np.array([a, b, b, b, b, b, b])

        tk = 0

        def fX(x, P):
            X = np.zeros([3, 7])
            X[:, 0] = x
            X[:, 1] = x + np.array([math.sqrt((nx + kappa) * P[0, 0]), 0, 0])
            X[:, 2] = x + np.array([0, math.sqrt((nx + kappa) * P[1, 1]), 0])
            X[:, 3] = x + np.array([0, 0, math.sqrt((nx + kappa) * P[2, 2])])
            X[:, 4] = x - np.array([math.sqrt((nx + kappa) * P[0, 0]), 0, 0])
            X[:, 5] = x - np.array([0, math.sqrt((nx + kappa) * P[1, 1]), 0])
            X[:, 6] = x - np.array([0, 0, math.sqrt((nx + kappa) * P[2, 2])])
            return X

        def fa(X):
            for i in range(7):
                X[:, i] = fx(X[:, i], 0)
            return X

        xhat = [2., .055, .044]
        Ptil = .01 * np.eye(3)

        def fXhat(X, Rww):
            Xhat = np.zeros([3, 7])
            Xhat[:, 0] = X[:, 0]
            Xhat[:, 1] = X[:, 1] + np.array([kappa * math.sqrt(Rww[0, 0]), 0, 0])
            Xhat[:, 2] = X[:, 2] + np.array([0, kappa * math.sqrt(Rww[1, 1]), 0])
            Xhat[:, 3] = X[:, 3] + np.array([0, 0, kappa * math.sqrt(Rww[2, 2])])
            Xhat[:, 4] = X[:, 4] - np.array([kappa * math.sqrt(Rww[0, 0]), 0, 0])
            Xhat[:, 5] = X[:, 5] - np.array([0, kappa * math.sqrt(Rww[1, 1]), 0])
            Xhat[:, 6] = X[:, 6] - np.array([0, 0, kappa * math.sqrt(Rww[2, 2])])
            return Xhat

        def fc(Xhat):
            Y = np.zeros(7)
            for i in range(7):
                Y[i] = fy(Xhat[:, i], 0)
            return Y

        X = fX(xhat, Ptil)
        Xhat = fXhat(X, Rww)
        Y = fc(Xhat)
        yhat = W @ Y
        e = yts[tk] - yhat
        ep = Y - yhat
        Repep = ep @ np.diag(W) @ ep.T + Rvv

        xhatts = np.zeros([n, 3])
        xtilts = np.zeros([n, 3])
        yhatts = np.zeros(n)
        ets = np.zeros(n)
        for tk in range(1, len(tts)):
            X = fa(fX(xhat, Ptil))
            xhat = W @ X.T
            Xtil = (X.T - xhat).T
            Ptil = Xtil @ np.diag(W) @ Xtil.T + Rww
            X = fX(xhat, Ptil)
            Xhat = fXhat(X, Rww)
            Y = fc(Xhat)
            yhat = W @ Y
            e = yts[tk] - yhat
            ep = Y - yhat
            Repep = ep @ np.diag(W) @ ep.T + Rvv
            Rxtilep = Xtil @ np.diag(W) @ ep
            K = Rxtilep / Repep
            # xhat = xhat + K * e
            # Ptil = Ptil - K @ (Repep * K.T)

            xhatts[tk, :] = xhat
            xtilts[tk, :] = xts[tk] - xhat
            yhatts[tk] = yhat
            ets[tk] = e
        innov = Innov(tts, ets)
        innov.abp(tts, xhatts, xtilts, yhatts)

if __name__ == "__main__":
    mod = Modern()
    mod.spbp1()
    mod.spbp2()



import numpy as np
from innov import Innov

class Classical():
    def __init__(self):
        self.log = []

    def sim1_linearized(self):
        from sim import Sim1
        sim = Sim1()
        dt = sim.dt
        def c(x): return x**2 + x**3
        def A(x): return 1 - .05*dt + .08*dt*x
        def C(x): return 2*x + 3*x**2
        xref, xhat, Ptil = 2., 2.2, .01
        for step in sim.steps():
            xhat = ((1-.05*dt)*xref + .04*dt*xref**2) + A(xref) * (xhat - xref)
            Ptil = A(xref)**2 * Ptil
            xref = 2. + .067 * step[0]
            Ree = C(xhat)**2*Ptil + sim.Rvv
            K = Ptil*C(xref)/Ree
            yhat = (xref**2 + xref**3) + C(xref) * (xhat - xref)
            xhat = xhat + K*(step[2]-yhat)
            Ptil = (1 - K*C(xref)) * Ptil
            self.log.append([step[0], xhat, yhat, step[1]-xhat, step[2]-yhat])
        innov = Innov(self.log)
        innov.plot_standard()

    def sim1_extended(self):
        from sim import Sim1
        sim = Sim1()
        dt = sim.dt
        def a(x): return (1 - .05*dt)*x + (.04*dt)*x**2
        def c(x): return x**2 + x**3
        def A(x): return 1 - .05*dt + .08*dt*x
        def C(x): return 2*x + 3*x**2
        xref, xhat, Ptil = 2., 2.2, .01
        for step in sim.steps():
            Phi = A(a(xhat))
            xhat = a(xhat)
            Ptil = Phi*Ptil*Phi
            yhat = c(xhat)
            Ree = C(xhat)*Ptil*C(xhat) + sim.Rvv
            K = Ptil*C(xhat)/Ree
            xhat = xhat + K*(step[2]-yhat)
            Ptil = (1 - K*C(xhat))*Ptil
            self.log.append([step[0], xhat, yhat, step[1]-xhat, step[2]-yhat])
        innov = Innov(self.log)
        innov.plot_standard()

    def sim1b_adaptive(self):
        from sim import Sim1b
        sim = Sim1b()
        dt = sim.dt
        def a(x): return np.array([(1 - x[1]*dt)*x[0] + x[2]*dt*x[0]**2, x[1], x[2]])
        def c(x): return x[0]**2 + x[0]**3
        def A(x):
            A = np.eye(3)
            A[0, 0] = 1 - x[1]*dt + 2*x[2]*dt*x[0]
            A[0, 1] = -dt*x[0]
            A[0, 2] = dt*x[0]**2
            return A
        def C(x): return np.array([2*x[0] + 3*x[0]**2, 0, 0])
        xhat = [2.2, .055, .044]
        Ptil = 100. * np.eye(3)
        U, D = Classical.UD(Ptil)
        for step in sim.steps():
            Phi = A(a(xhat))
            yhat = c(xhat)
            xhat, U, D = Classical.thornton(xin=xhat, Phi=Phi, Uin=U, Din=D, Gin=np.eye(3), Q=sim.Rww)
            xhat, U, D = Classical.bierman(z=step[2]-yhat, R=sim.Rvv, H=C(xhat), xin=xhat, Uin=U, Din=D)
            self.log.append([step[0], xhat[0], yhat, step[1][0]-xhat[0], step[2]-yhat])
        innov = Innov(self.log)
        innov.plot_standard()

    @staticmethod
    def UD(M):
        assert np.allclose(M, M.T)
        n = M.shape[0]
        M = np.triu(M)
        U = np.eye(n)
        d = np.zeros(n)
        for j in reversed(range(2, n + 1)):
            d[j - 1] = M[j - 1, j - 1]
            if d[j - 1] > 0:
                alpha = 1.0 / d[j - 1]
            else:
                alpha = 0.0
            for k in range(1, j):
                beta = M[k - 1, j - 1]
                U[k - 1, j - 1] = alpha * beta
                M[0:k, k - 1] = M[0:k, k - 1] - beta * U[0:k, j - 1]
        d[0] = M[0, 0]
        return U, np.diag(d)

    @staticmethod
    def thornton(xin, Phi, Uin, Din, Gin, Q):
        x, U, D = Phi @ xin, Uin, Din
        n, r = 3, 3
        G = Gin
        U = np.eye(3)
        PhiU = Phi @ Uin
        for i in reversed(range(3)):
            sigma = 0
            for j in range(n):
                sigma = sigma + PhiU[i,j]**2 * Din[j,j]
                if (j <= r-1):
                    sigma = sigma + G[i,j]**2 + Q[j,j]
            D[i,i] = sigma
            ilim = i-1
            if not ilim < 0:
                for j in range(ilim):
                    sigma = 0
                    for k in range(n):
                        sigma = sigma + PhiU[i,k] * Din[k,k] * PhiU[j,k]
                    for k in range(r):
                        sigma = sigma + G[i,k] * Q[k,k] * G[j,k]
                    U[j,i] = sigma / D[i,i]
                    for k in range(n):
                        PhiU[j,k] = PhiU[j,k] - U[j,i] * PhiU[i,k]
                    for k in range(r):
                        G[j,k] = G[j,k] - U[j,i] * G[i,k]
        return x, U, D

    @staticmethod
    def bierman(z, R, H, xin, Uin, Din):
        x, U, D = xin, Uin, Din
        a = U.T @ H.T
        b = D @ a
        dz = z # z - H @ xin
        alpha = R
        gamma = 1 / alpha
        for j in range(3):
            beta = alpha
            alpha = alpha + a[j] * b[j]
            lamda = -a[j] * gamma
            gamma = 1 / alpha
            D[j, j] = beta * gamma * D[j, j]
            jlim = j-1
            if not jlim < 0:
                for i in range(jlim):
                    beta = U[i, j]
                    U[i, j] = beta + b[i] * lamda
                    b[i] = b[i] + b[j] * beta
        dzs = gamma * dz
        x = x + dzs * b
        return x, U, D

if __name__ == "__main__":
    cl = Classical()
    cl.sim1_linearized()
    cl = Classical()
    cl.sim1_extended()
    cl = Classical()
    cl.sim1b_adaptive()



# monte carlo sampling processor, bootstrap particle filter
import numpy as np
import math
from innov import Innov

class Particle():
    def __init__(self):
        pass

    def pf1(self):
        nsamp = 250
        n = 150
        deltat = .01

        tts = np.arange(0, n * deltat, deltat)
        Rww = 1e-6
        wts = math.sqrt(Rww) * np.random.randn(n)
        Rvv = 9e-2
        vts = math.sqrt(Rvv) * np.random.randn(n)

        def fx(x, w):
            return (1 - .05 * deltat) * x + .04 * deltat * x**2 + w

        def fy(x, v):
            return x ** 2 + x ** 3 + v

        vfA = np.vectorize(fx)

        def fC(y, xi):
            return np.exp(-np.log(2. * np.pi * Rvv) / 2. - (y - xi**2 - xi**3)**2 / (2. * Rvv))
        vfC = np.vectorize(fC)

        xts = np.zeros((n,))
        xts[0] = 2.
        yts = np.zeros((n,))
        yts[0] = fy(xts[0], vts[0])
        for tk in range(1, n):
            xts[tk] = fx(xts[tk - 1], wts[tk - 1])
            yts[tk] = fy(xts[tk], vts[tk])

        xhatts = np.zeros((n,))
        xhatts[0] = 2.2
        xtilts = np.zeros((n,))
        xtilts[0] = xts[0] - xhatts[0]

        wits = math.sqrt(Rww) * np.random.randn(n, nsamp)
        xits = np.zeros((n, nsamp))
        xits[0, :] = xhatts[0] + wits[0, :]

        Wits = np.zeros((n, nsamp))
        Wi = vfC(yts[0], xits[0, :])
        Wits[0, :] = Wi / sum(Wi)

        yhatts = np.zeros((n,))
        yhatts[0] = fy(xhatts[0], 0)
        ets = np.zeros((n,))
        ets[0] = yts[0] - yhatts[0]

        for tk in range(1, n):
            xits[tk, :] = vfA(xits[tk - 1, :], wits[tk, :])
            Wi = vfC(yts[tk], xits[tk, :])
            Wits[tk, :] = Wi / sum(Wi)
            xhatts[tk] = Wits[tk, :] @ xits[tk, :]
            xtilts[tk] =  xts[tk] - xhatts[tk]
            yhatts[tk] = fy(xhatts[tk], 0)
            ets[tk] = yts[tk] - yhatts[tk]
            xits[tk, :] = Particle.resample(xits[tk, :], Wits[tk, :])

        innov = Innov(tts, ets)
        innov.plot_standard(tts, xhatts, xtilts, yhatts)

    @staticmethod
    def resample(xi, Wi):
        tmp = []
        for i in range(xi.size):
            tmp.append([xi[i], Wi[i]])
        tmp = sorted(tmp, key=lambda x: x[0])
        cdf = [[tmp[0][0], tmp[0][1]]]
        for i in range(1, len(tmp)):
            cdf.append([tmp[i][0], tmp[i][1] + cdf[i-1][1]])
        cdf = np.asarray(cdf)
        uk = np.sort(np.random.uniform(size=xi.size))
        xhati, k = [], 0
        for row in cdf:
            while k < uk.size and uk[k] <= row[1]:
                xhati.append(row[0])
                k += 1
        xhati = np.asarray(xhati)
        assert xhati.size == xi.size
        return xhati

if __name__ == "__main__":
    pf = Particle()
    pf.pf1()



import math
import numpy as np

class Sim1():
    def __init__(self):
        self.tsteps = 151
        self.dt = .01
        self.x = 2.
        self.Rww = .000001
        self.Rvv = .09
        self.log = []

    def steps(self):
        for tstep in range(self.tsteps):
            tsec = tstep * self.dt
            w = math.sqrt(self.Rww) * np.random.randn()
            v = math.sqrt(self.Rvv) * np.random.randn()
            self.x = (1 - .05 * self.dt) * self.x + (.04 * self.dt) * self.x**2 + w
            self.y = self.x**2 + self.x**3 + v
            if tstep == 0: continue
            self.log.append([tsec, self.x, self.y])
            yield (tsec, self.x, self.y)

class Sim1b():
    def __init__(self):
        self.tsteps = 151
        self.dt = .01
        self.x = np.array([2., .05, .04])
        self.Rww = np.diag([0, 0, 0])
        self.Rvv = .09
        self.log = []

    def steps(self):
        for tstep in range(self.tsteps):
            tsec = tstep * self.dt
            w = np.multiply(np.random.randn(1, 3), np.sqrt(np.diag(self.Rww)))
            v = math.sqrt(self.Rvv) * np.random.randn()
            self.x = np.array([(1 - self.x[1]*self.dt)*self.x[0] + self.x[2]*self.dt*self.x[0]**2, self.x[1], self.x[2]]) + w[0]
            self.y = self.x[0]**2 + self.x[0]**3 + v
            if tstep == 0: continue
            self.log.append([tsec, self.x, self.y])
            yield (tsec, self.x, self.y)

if __name__ == "__main__":
    sim = Sim1()
    for step in sim.steps():
        print(step)

